//     File name: bump.c
//        Author: Gordon Griesel   <73042,2246>
//          Date: 02/27/95
//      Compiler: Borland C++ 3.0
//Development PC: 486 DX2-66 VESA
//
// (C) Copyright Gordon Griesel 1995
//
//--------------------------------------------------------------------------
//Author's notes:
//
//This program demonstrates my method for handling billiard-ball type
//collisions with objects of different masses.
//
//I was inspired to write this program after seeing Dave Bollinger's CBENCH.
//My goal was to duplicate his animation, but start from scratch.  Then,
//I might come up with some new way to get the job done.
//
//This program leaves plenty of room for optimization, but I believe
//that in its present form, you can learn something about collision physics.
//A situation that I do not handle is when one object is struck by
//two other objects at the same time.  If this happens, you may see an
//object pass through another object.
//
//Please don't be shy about making comments.  I learn from my mistakes,
//and I'll bet I made some in this code.  If you find errors
//or think of any optimizations, let me know.
//
//--------------------------------------------------------------------------
//
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <dos.h>
//
#define PI            3.14159265358979
#define PId2          1.5707963279489
#define MAX_OBJECTS  19
//
void ShowAnimation(int duration);
void SetBounceFlags(int i, int j);
void SetDAC(int col, unsigned char red, unsigned char green, unsigned char blue);
int  CheckForKeystroke(void);
void Circle(int xc, int yc, int radius, int color);
//
union REGS regs;
char bounceFlag[MAX_OBJECTS+1][MAX_OBJECTS+1];
int nObjects;
int scenario=4;
char far *ScreenPtr;


void main(void)
{
	unsigned int sSegment, sOffset;

	//set mode 13h
	regs.h.ah = 0x00;
	regs.h.al = 0x13;
	int86(0x10, &regs, &regs);
	//
	//setup a pointer to screen mem
	sSegment = 0x0a000;
	sOffset  = 0;
	ScreenPtr = MK_FP(sSegment, sOffset);
	randomize();
	do
	{
		ShowAnimation(600);
		if (CheckForKeystroke()) break;
	} while(1);
	//
	//cleanup
	if (kbhit())
		if (getch() == 0) getch();
	regs.h.ah = 0x00;
	regs.h.al = 0x03;
	int86(0x10, &regs, &regs);
}

void ShowAnimation(int duration)
{
	int i,j,c1;
	int ix,iy;
	int oldx[MAX_OBJECTS+1], oldy[MAX_OBJECTS+1];
	int spec[MAX_OBJECTS+1];
	unsigned char oColor[MAX_OBJECTS+1];
	float gravity;
	float x[MAX_OBJECTS+1], y[MAX_OBJECTS+1];
	float xmove[MAX_OBJECTS+1], ymove[MAX_OBJECTS+1];
	float radius[MAX_OBJECTS+1];
	float mass[MAX_OBJECTS+1];
	float contactySpeed, contactxSpeed;
	float distance, speed;
	float tx, ty;
	float contactAngle;
	float closingAngle;
	float xclosingjtoi, yclosingjtoi;
	float adiff, portion;
	float massFactor, massi, massj;

	//Initial setup of objects.
	//-------------------------------
	//
	scenario = (++scenario % 5);
	switch(scenario)
	{
		case 0:
		{
			int xpos[14] = {40,40,40,40,40,40,40,64,64,64,64,64,64,64};
			int k;
			nObjects = 14;
			gravity = 0.0;
				  x[0] = 300;
				  y[0] = (float)random(160)+20;
			 xmove[0] = -((float)random(1000)/1000 * 3 + 2);
			 ymove[0] = (float)random(1000)/1000 * 2 - 1;
			radius[0] = 4;
			  mass[0] = radius[0] * PI;
			  spec[0] = (int)(radius[0]/1.8);
			j = 1;
			k = 0;
			for (i=0; i<14; i++)
			{
					  x[j] = (float)xpos[i];
					  y[j] = (++k % 7)*24 + 24;
				 xmove[j] = 0.0;
				 ymove[j] = 0.0;
				radius[j] = 10;
				  mass[j] = radius[j] * PI;
				  spec[j] = (int)(radius[j]/1.8);
				  j++;
			}
			break;
		}
		case 1:
		{
			int xpos[8] = {40,40,40,40,64,64,64,64};
			int ypos[8] = {0,1,2,3,0,1,2,3};
			nObjects = 8;
			gravity = 0.04;
			duration = 1200;
				  x[0] = 280;
				  y[0] = (float)random(60)+120;
			 xmove[0] = -((float)random(1000)/1000 + .5);
			 ymove[0] = (float)random(1000)/1000 * 2 - 1;
			radius[0] = 4;
			  mass[0] = radius[0] * PI;
			  spec[0] = (int)(radius[0]/1.8);
			j = 1;
			for (i=0; i<8; i++)
			{
					  x[j] = (float)xpos[i];
					  y[j] = (float)ypos[i]*42 + 20;
				 xmove[j] = 0.0;
				 ymove[j] = 0.0;
				radius[j] = 10;
				  mass[j] = radius[j] * PI;
				  spec[j] = (int)(radius[j]/1.8);
				  j++;
			}
			break;
		}
		case 2:
		{
			int xy[8] = {60,60,40,140,140,120,100,160};
			int rad[4] = {50,7,12,16};
			//
			nObjects = 4;
			gravity = 0.0;
			duration = 1200;
				  x[0] = 300;
				  y[0] = random(160)+20;
			 xmove[0] = -((float)random(1000)/1000 * 5 + 3);
			 ymove[0] = (float)random(1000)/1000 * 2 - 1;
			radius[0] = 4;
			  mass[0] = radius[0] * PI;
			  spec[0] = (int)(radius[0]/1.8);
			j = 1;
			for (i=0; i<4; i++)
			{
					  x[j] = (float)xy[i*2];
					  y[j] = (float)xy[i*2+1];
				 xmove[j] = 0.0;
				 ymove[j] = 0.0;
				radius[j] = (float)rad[i];
				  mass[j] = radius[j] * PI;
				  spec[j] = (int)(radius[j]/1.8);
				  j++;
			}
			break;
		}
		case 3:
		{
			int ypos[4] = {16,40,74,154};
			int rad[4] = {4,10,14,40};
			//
			nObjects = 3;
			gravity = 0.04;
			duration = 1400;
			for (i=0; i<4; i++)
			{
					  x[i] = (float)160;
					  y[i] = (float)ypos[i];
				 xmove[i] = (float)random(1000)/1000 * .0001;
				 ymove[i] = 0.0;
				radius[i] = (float)rad[i];
				  mass[i] = radius[i] * PI;
				  spec[i] = (int)(radius[i]/1.8);
			}
			break;
		}
		case 4:
		{
			int xpos[9] = {160,140,140,140,140,160,160,160,160};
			int ypos[9] = {156,20,40,60,80,20,40,60,80};
			int rad[9] = {40,4,4,4,4,4,4,4,4};
			//
			nObjects = 8;
			gravity = 0.04;
			duration = 1200;
			for (i=0; i<9; i++)
			{
					  x[i] = (float)xpos[i] + (float)random(1000)/1000 * .4 - .2;
					  y[i] = (float)ypos[i];
				 xmove[i] = (float)random(1000)/1000 * .001;
				 ymove[i] = 0.0;
				radius[i] = (float)rad[i];
				  mass[i] = radius[i] * PI;
				  spec[i] = (int)(radius[i]/1.8);
			}
			break;
		}
	}
	//Clear the screen
	for (i=0; i<200; i++)
	{
		c1 = random(100);
		for (j=0; j<320; j++)
			*(ScreenPtr+(i*320)+j) = 0;
	}
	for (i=0; i<=nObjects; i++)
	{
		oColor[i] = i+20;
		SetDAC(oColor[i], (unsigned char)random(54)+10,
								(unsigned char)random(54)+10,
								(unsigned char)random(54)+10 );
	}
	for (i=0; i<=nObjects; i++)
		for (j=0; j<=nObjects; j++)
			bounceFlag[i][j] = 0;
	//
	//-------------------
	//Main animation loop
	//-------------------
	do
	{
		for (i=0; i<=nObjects; i++)
		{
			for (j=i+1; j<=nObjects; j++)
			{
				tx = x[i] - x[j];
				ty = y[i] - y[j];
				distance = (tx * tx + ty * ty);
				if (distance <= (radius[i]+radius[j])*(radius[i]+radius[j]))
				{
					//We have a collision!
					if (bounceFlag[i][j] == 0)
					{
						//Determine the angle of contact
						if (tx != 0)
						{
							contactAngle = atan(ty/tx);
							if (contactAngle < 0) contactAngle = -contactAngle;
						}
						else
							contactAngle = PId2;
						//To determine how the objects will rebound off of each
						//other, we are not concerned with the speed of each
						//object, but rather the relative, or closing, speeds
						//of the two objects.
						xclosingjtoi = xmove[j] - xmove[i];
						yclosingjtoi = ymove[j] - ymove[i];
						//Calculate the x & y speed in the direction of contact
						speed = sqrt(yclosingjtoi*yclosingjtoi+xclosingjtoi*xclosingjtoi);
						contactySpeed = sin(contactAngle) * speed;
						contactxSpeed = cos(contactAngle) * speed;
						//Now, determine the closing angle.
						if (xclosingjtoi != 0)
						{
							closingAngle = atan(yclosingjtoi/xclosingjtoi);
							if (closingAngle < 0) closingAngle = -closingAngle;
						}
						else
							closingAngle = PId2;
						//Hmmmm...
						//With equal masses,
						//   the max rebound speed is 1/2 the closing speed.
						//With unequal masses,
						//   the max rebound speed is the closing speed.
						//Okay then,
						//Normalize the two masses to be:  mass1 + mass2 = 2.0
						massFactor = 2 / (mass[i] + mass[j]);
						massi = mass[i] * massFactor;
						massj = mass[j] * massFactor;
						//Quadrant-specific stuff
						if (x[i] > x[j])
						{
							if (y[i] < y[j])
							{
								//[i] is contacting upper right quadrant of [j]
								if (yclosingjtoi < 0)
								{
									if (xclosingjtoi < 0)
										adiff = PI - contactAngle - closingAngle;
									else
										adiff = contactAngle - closingAngle;
								}
								else
									adiff = contactAngle + closingAngle;
								portion = cos(adiff);
								xmove[i] += contactxSpeed * portion * massj;
								ymove[i] -= contactySpeed * portion * massj;
								xmove[j] -= contactxSpeed * portion * massi;
								ymove[j] += contactySpeed * portion * massi;
							}
							else
							{
								//[i] is contacting lower right quadrant of [j]
								if (yclosingjtoi > 0)
								{
									if (xclosingjtoi < 0)
										adiff = PI - contactAngle - closingAngle;
									else
										adiff = contactAngle - closingAngle;
								}
								else
									adiff = contactAngle + closingAngle;
								portion = cos(adiff);
								xmove[i] += contactxSpeed * portion * massj;
								ymove[i] += contactySpeed * portion * massj;
								xmove[j] -= contactxSpeed * portion * massi;
								ymove[j] -= contactySpeed * portion * massi;
							}
						}
						else
						{
							if (y[i] < y[j])
							{
								//[i] is contacting upper left quadrant of [j]
								if (yclosingjtoi < 0)
								{
									if (xclosingjtoi < 0)
										adiff = contactAngle - closingAngle;
									else
										adiff = PI - contactAngle - closingAngle;
								}
								else
									adiff = contactAngle + closingAngle;
								portion = cos(adiff);
								xmove[i] -= contactxSpeed * portion * massj;
								ymove[i] -= contactySpeed * portion * massj;
								xmove[j] += contactxSpeed * portion * massi;
								ymove[j] += contactySpeed * portion * massi;
							}
							else
							{
								//[i] is contacting lower left quadrant of [j]
								if (yclosingjtoi > 0)
								{
									if (xclosingjtoi < 0)
										adiff = contactAngle - closingAngle;
									else
										adiff = PI - contactAngle - closingAngle;
								}
								else
									adiff = contactAngle + closingAngle;
								portion = cos(adiff);
								xmove[i] -= contactxSpeed * portion * massj;
								ymove[i] += contactySpeed * portion * massj;
								xmove[j] += contactxSpeed * portion * massi;
								ymove[j] -= contactySpeed * portion * massi;
							}
						}
						SetBounceFlags(i,j);
					}
				}
				else
					bounceFlag[i][j] = 0;
			}
			Circle(oldx[i], oldy[i], radius[i], 0);
			c1 = spec[i];
			*(ScreenPtr+((oldy[i]-c1)*320) + oldx[i]-c1) = 0;
			*(ScreenPtr+((oldy[i]-c1-1)*320) + oldx[i]-c1+1) = 0;
			*(ScreenPtr+((oldy[i]-c1+1)*320) + oldx[i]-c1-1) = 0;
			x[i] += xmove[i];
			y[i] += ymove[i];
			ymove[i] += gravity;
			if (x[i] < (radius[i]+4) || x[i] > 315-radius[i])
			{
				xmove[i] = -xmove[i];
				x[i] += (xmove[i] + xmove[i]);
				SetBounceFlags(i,i);
			}
			if (y[i] <= (radius[i]+4) || y[i] > 195-radius[i])
			{
				ymove[i] = -ymove[i];
				if (gravity) ymove[i] *= .8;
				y[i] += (ymove[i] + ymove[i]);
				SetBounceFlags(i,i);
			}
			ix = (int)x[i];
			iy = (int)y[i];
			Circle(ix, iy, radius[i], oColor[i]);
			oldx[i] = ix;
			oldy[i] = iy;
			*(ScreenPtr+((iy-c1)*320) + ix-c1) = 15;
			*(ScreenPtr+((oldy[i]-c1-1)*320) + oldx[i]-c1+1) = 7;
			*(ScreenPtr+((oldy[i]-c1+1)*320) + oldx[i]-c1-1) = 7;
		}
		do { if ((inp(0x3da)&8) == 8) break; } while(1);
		if (kbhit()) break;
		if (--duration < 1) break;
	} while(1);
}

void SetBounceFlags(int i, int j)
{
	//When object-1 strikes object-2, we will calculate the rebound one
	//time, and then not again until at least one of the objects has
	//struck a wall or another object.
	//The rebound created from a collision may not completely seperate
	//the two objects in just one frame.
	int k;

	for (k=0; k<=nObjects; k++)
	{
		bounceFlag[k][j] = 0;
		bounceFlag[k][i] = 0;
		bounceFlag[i][k] = 0;
		bounceFlag[j][k] = 0;
	}
	bounceFlag[i][j] = 1;
}

void SetDAC(int col, unsigned char red, unsigned char green, unsigned char blue)
{
	regs.h.ah = 0x10; //BOIS service
	regs.h.al = 0x10; //function 10h
	regs.x.bx = col;
	regs.h.dh = red;
	regs.h.ch = green;
	regs.h.cl = blue;
	int86(0x10, &regs, &regs);
}

int CheckForKeystroke(void)
{
	int xkey;

	if (kbhit())
	{
		xkey = getch();
		if (xkey == 0)
		{
			getch();
			return 1;
		}
		else
			if (xkey == 'n' || xkey == 'N')
				return 0;
			else
				return 1;
	}
	else
		return 0;
}

void Circle(int xc, int yc, int radius, int color)
{
	//This code is adapted from a Richard Ferraro book entitled
	//Programmer's Guide to the EGA & VGA Cards. (pg 535).
	int x=0, y, d;
	int xxcp,xxcm,xycp,xycm,yxcp,yxcm,yycp,yycm;

	y = radius;
	d = 3 - (radius << 1);
	do
	{
		xxcp = xc+x; xxcm = xc-x; xycp = xc+y;
		xycm = xc-y; yxcp = yc+x; yxcm = yc-x;
		yycp = yc+y; yycm = yc-y;
		*(ScreenPtr+(yycp*320)+xxcp) = color;
		*(ScreenPtr+(yycm*320)+xxcp) = color;
		*(ScreenPtr+(yycp*320)+xxcm) = color;
		*(ScreenPtr+(yycm*320)+xxcm) = color;
		*(ScreenPtr+(yxcp*320)+xycp) = color;
		*(ScreenPtr+(yxcm*320)+xycp) = color;
		*(ScreenPtr+(yxcp*320)+xycm) = color;
		*(ScreenPtr+(yxcm*320)+xycm) = color;
		if (d < 0)
			d += ( (x << 2) + 6);
		else
			d += (((x - y--) << 2) + 10);
		if (++x > y) break;
	} while(1);
}
